"use strict";

const express = require('express')
const bodyParser = require('body-parser')
const fileUpload = require('express-fileupload');
const constants = require('.././constants.json')
const fs = require('fs')
const { spawn } = require('child_process');
const morgan = require('morgan')

const app = express()
const libSupport = require('./lib')
const logger = libSupport.logger
let date = new Date();
let log_channel = constants.log_channel

let functionToResource = new Map(), // TODO: make the resource a list for horizontal scale out
    usedPort = new Map(), // TODO: remove after integration with RM
    rmQueue = new Map(), // queue holding requests for which DM is waiting for RM allocation
    db = new Map(),
    resourceMap = new Map()

let kafka = require('kafka-node'),
    Producer = kafka.Producer,
    client = new kafka.KafkaClient({ 
        kafkaHost: constants.kafka_host,
        autoConnect: true
    }),
    producer = new Producer(client),
    Consumer = kafka.Consumer,
    consumer = new Consumer(client,
        [
            { topic: 'response', partition: 0, offset: 0}, 
            { topic: 'heartbeat' }, 
            { topic: "deployed" },
            { topic: "removeWorker" },
            { topic: "RESPONSE_RM_2_DM"}
        ],
        [
            { autoCommit: true }
        ])

app.use(morgan('combined'))
app.use(bodyParser.urlencoded({ extended: true }))
app.use(bodyParser.json())
const file_path = __dirname + "/repository/"

app.use('/repository', express.static(file_path));
app.use(fileUpload())

let requestQueue = []
let workerNodes = []

const WINDOW_SIZE = 10
const port = constants.master_port
const registry_url = constants.registry_url

/**
 * REST API to receive deployment requests
 */
app.post('/serverless/deploy', (req, res) => {
    
    let runtime = req.body.runtime
    let file = req.files.serverless

    let functionHash = file.md5
    
    file.mv(file_path + functionHash, function (err) {
        functionHash = libSupport.generateExecutor(file_path, functionHash)
        if (err) {
            logger.error(err)
            res.send("error").status(400)
        }
        else {
            if (runtime === "container") {
                deployContainer(file_path, functionHash)
                    .then(() => {
                        res.json({
                            status: "success",
                            function_id: functionHash
                        })
                    })
                    .catch(err => {
                        res.json({
                            status: "error",
                            reason: err}).status(400)
                    })
            } else {
                res.json({
                    status: "success",
                    function_id: functionHash
                })
            }
        }
    })
    
})

/**
 * Create the docker file, build and push image to remote repository
 * @param {String: Path from where to extract function executor} path 
 * @param {String: Name of the image} imageName 
 */
function deployContainer(path, imageName) {
    return new Promise((resolve, reject) => {
        let buildStart = Date.now()
        
        fs.writeFile('./repository/Dockerfile',
            `FROM node:latest
            WORKDIR /app
            COPY package.json /app
            RUN npm install
            COPY . /app
            ENTRYPOINT ["node", "${imageName}.js"]`
            , function (err) {
                if (err) {
                    logger.error("failed", err);

                    
                    reject(err);
                }
                else {
                    logger.info('Dockerfile created');
                    const process = spawn('docker', ["build", "-t", registry_url + imageName, path]);

                    process.stdout.on('data', (data) => {
                        logger.info(`stdout: ${data}`);

                    });

                    process.stderr.on('data', (data) => {
                        logger.error(`stderr: ${data}`);
                    });

                    process.on('close', (code) => {
                        logger.warn(`child process exited with code ${code}`);
                        let timeDifference = Math.ceil((Date.now() - buildStart))
                        logger.info("image build time taken: ", timeDifference);
                        const process_push = spawn('docker', ["push", registry_url + imageName]);

                        process_push.stdout.on('data', (data) => {
                            console.log(`stdout: ${data}`);

                        });

                        process_push.stderr.on('data', (data) => {
                            logger.error(`stderr: ${data}`);
                        });

                        process_push.on('close', (code) => {
                            logger.info("image pushed to repository");
                            resolve();
                        })
                        
                    });
                }
            });
    })
}

/**
 * REST API to receive execute requests
 */
app.post('/serverless/execute/:id', (req, res) => {
    let runtime = req.body.runtime
    if (functionToResource.has(req.params.id + runtime)) {
        /**
         * Bypass deployment pipeline if resource available
         */
        let forwardTo = functionToResource.get(req.params.id + runtime) 
        let resource = resourceMap.get(forwardTo.resource_id)
        logger.info("resource found " + JSON.stringify(forwardTo) + 
            " forwarding via reverse proxy to: " + JSON.stringify(resource));
        libSupport.reverseProxy(req, res, `http://${resource.node_id}:${resource.port}/serverless/function/execute`)
        
    } else {
        /**
         * FIXME: Here, every request even for the same function will be queued up potentially launching multiple 
         * resource of the same type
         */
        requestQueue.push({ req, res })
        /**
         * We store functions for function placement heuristics purposes. This lets us look into the function
         * patterns being received and make intelligent deployment decisions based on it.
         */
        if (requestQueue.length >= WINDOW_SIZE)
            dispatch()
    }
})

/**
 * Send dispatch signal and deploy resources after consultation with the RM
 */
function dispatch() {
    let lookbackWindow = Math.min(WINDOW_SIZE, requestQueue.length)
    for (let i = 0; i < lookbackWindow; i++) {
        let {req, res} = requestQueue.shift()
	    // logger.info(req.body)
        let runtime = req.body.runtime
        let functionHash = req.params.id
        
        let resource_id = libSupport.makeid(20) // each function resource request is associated with an unique ID
        logger.info(`Generated new resource ID: ${resource_id} for runtime: ${runtime}`);
        
        let node_id = getAddress() // Requests the RM for address and other metadata for function placement
        let port = libSupport.getPort(usedPort) // TODO: will be provided by the RM

        let payload = [{
            topic: node_id,
            messages: JSON.stringify({
                "type": "execute", // Request sent to Dispatch Daemon via Kafka for actual deployment at the Worker
                resource_id,
                runtime, functionHash,
                port
            }),
            partition: 0
        }]
        logger.info("Requesting RM " + JSON.stringify({
            resource_id,
            "memory": 332,
        }))

        /** uncomment when RM is unavailable */
        
        resourceMap.set(resource_id, {
            runtime, functionHash, port, node_id
        })
        logger.info(resourceMap);
        producer.send(payload, () => {
            logger.info(`Resource Deployment request sent to Dispatch Agent`)
         })
        db.set(resource_id, { req, res })
        

        
        /** uncomment when RM is available, TODO: also update resourceMap
        rmQueue.set(resource_id, payload)
        let payloadToRM = [{
            topic: "REQUEST_DM_2_RM",
                messages: JSON.stringify({
                    resource_id,
                    "memory": 332,
                }),
                partition: 0
            }]
        producer.send(payloadToRM, () => {
            db.set(functionHash + runtime, { req, res })
        })
         */
        
    }
}

function getAddress() {
    return workerNodes[Math.floor(Math.random() * workerNodes.length)];
}

consumer.on('message', function (message) {
    
    let topic = message.topic
        message = message.value
    if (topic === "response") {
        logger.info("response " + message);
        
        
    } else if (topic === "heartbeat") {
        message = JSON.parse(message)
        if (Date.now() - message.timestamp < 1000)
            if (workerNodes.indexOf(message.address) === -1)  {
                workerNodes.push(message.address)
                logger.warn("New worker discovered. Worker List: " + workerNodes)
            }
    } else if (topic == "deployed") {
        try {
            message = JSON.parse(message)
        } catch (e) {
            // process.exit(0)
        }
        
        logger.info("Deployed Resource: " + JSON.stringify(message));
        
        if (db.has(message.resource_id)) {
            let { req, res } = db.get(message.resource_id)
            functionToResource.set(message.functionHash + message.runtime, {
                resource_id: message.resource_id
            })
            let resource = resourceMap.get(message.resource_id)
            let confirmRM = [{
                topic: log_channel,
                messages: JSON.stringify({
                    resource_id: message.resource_id,
                    node_id: resource.node_id,
                    runtime: resource.runtime, 
                    function_id: resource.functionHash,
                    "reason": "deployment",
                    "status": true,
                    "timestamp": date.toISOString()
                }),
                partition: 0
            }]
            producer.send(confirmRM, () => {
                logger.info(`Confirmed RM for successful deployment resource_id: ${message.resource_id} deployment`)
            })
            logger.info("forwarding request via reverse proxy to: " + JSON.stringify(resource));
            libSupport.reverseProxy(req, res, 
                `http://${resource.node_id}:${resource.port}/serverless/function/execute`)
                .then(() => {
                    db.delete(message.resource_id)
                })
        }
    } else if (topic == "removeWorker") {
        logger.info("Worker blown: Removing Metadata " + message);
        try {
            message = JSON.parse(message)
        } catch(e) {
            // process.exit(0)
        }
        usedPort.delete(message.port)
        let resource = functionToResource.get(message.functionHash + message.runtime)
        functionToResource.delete(message.functionHash + message.runtime)
        if (resource != null)
            resourceMap.delete(resource.resource_id)
    } else if (topic == "RESPONSE_RM_2_DM") {
        
        logger.info("Response from RM: " + message);
        message = JSON.parse(message)
        
        let payload = rmQueue.get(message.id)
        if (payload != null) {
            payload[0].topic = message.nodes[0]
            logger.info(payload);
            /** get port and other resources */
            let resource = resourceMap.get(message.id)
            resource.node_id = message.nodes[0] // TODO: update this to message.nodes[0].node_id
            // resource.port = message.nodes[0].port TODO: update after RM supports port allocation
            resourceMap.set(message.id, resource)
            logger.info(resourceMap);
            producer.send(payload, () => { })
        } else {
            logger.error("something went wrong");
            
        }
        
    }
});

setInterval(dispatch, 2000);
app.listen(port, () => logger.info(`Server listening on port ${port}!`))